from .view import View
from .data_source import GtPhlowColumnedListDataSource
from .column import ListColumn

class ColumnedListView(View):
	def __init__(self):
		super().__init__()
		self.itemsCallback = lambda : []
		self.columns = []
		self.accessor = None

	def items(self, itemsCallback):
		self.itemsCallback = itemsCallback
		return self
	
	def column(self, columnTitle, columnFormatCallback, columnWidth = None):
		tableColumn = ListColumn(columnTitle, columnFormatCallback, columnWidth)
		self.columns.append(tableColumn)
		return self

	def dataSource(self):
		return GtPhlowColumnedListDataSource(self.itemsCallback, self.columns, self.accessor)
	
	def asDictionaryForExport(self):
		exportData = super().asDictionaryForExport()
		exportData["viewName"] = "GtPhlowColumnedListViewSpecification"
		exportData["dataTransport"] = 2
		exportData["columnSpecifications"] = list(map(lambda column : {"title": column.getTitle(), "cellWidth": column.width, "type": "text", "spawnsObjects": False, "properties": []}, self.columns))
		return exportData
	


import bisect
import functools
import time
import inspect

from typing import Any
from abc import ABC, abstractmethod
from copy import copy

from gtoolkit_bridge import gtView

def methodevent(message=""):
    def decorate(func):
        @functools.wraps(func)
        def wrapped_function(*args, **kwargs):
            if 'signals' not in globals():
                return func(*args, **kwargs)
            nonlocal message
            if message=="":
                message = func.__name__
            start_signal = MethodStartSignal(message)
            try:
                start_signal.file = inspect.getsourcefile(func)
                [_, start_signal.line] = inspect.getsourcelines(func)
            except:
                pass
            try:
                value = func(*args, **kwargs)
                return value
            finally:
                end_signal = MethodEndSignal(message)
                try:
                   end_signal.file = inspect.getsourcefile(func)
                   [_, end_signal.line] = inspect.getsourcelines(func)
                except:
                   pass
        return wrapped_function
    return decorate

def argmethodevent(message=""):
    def decorate(func):
        @functools.wraps(func)
        def wrapped_function(*args, **kwargs):
            if 'signals' not in globals():
                return func(*args, **kwargs)
            nonlocal message
            if message=="":
                message = func.__name__           
            start_signal = ArgumentMethodStartSignal(message, args, kwargs)
            try:
                start_signal.file = inspect.getsourcefile(func)
                [_, start_signal.line] = inspect.getsourcelines(func)
            except:
                pass
            try:
                value = func(*args, **kwargs)
                return value
            finally:
                end_signal = MethodEndSignal(message)
                try:
                    end_signal.file = inspect.getsourcefile(func)
                    [_, end_signal.line] = inspect.getsourcelines(func)
                except:
                    pass
        return wrapped_function
    return decorate

def gtTrace(func):
    @functools.wraps(func)
    def wrapper_gtTrace(*args, **kwargs):
        if 'signals' not in globals():
            return func(*args, **kwargs)
        funcName = func.__name__
        funcArgNames = inspect.getfullargspec(func).args
        funcArgs = [(name, copy(value)) for name, value in zip(funcArgNames, args)]
        if (funcArgNames[0] == 'self'):
            receiver = args[0]
            funcArgs = funcArgs[1:]
        else:
            receiver = None
        start_signal = ArgumentMethodStartSignal(receiver, funcName, funcArgs, kwargs)
        try:
            start_signal.file = inspect.getsourcefile(func)
            [_, start_signal.line] = inspect.getsourcelines(func)
        except:
            pass
        value = func(*args, **kwargs)
        end_signal = ResultMethodEndSignal(receiver, funcName, value)
        try:
            end_signal.file = inspect.getsourcefile(func)
            [_, end_signal.line] = inspect.getsourcelines(func)
        except:
            pass
        return value
    return wrapper_gtTrace

class Telemetry(ABC):
    def __init__(self, message):
        super().__init__
        self.message = message

    def children(self):
        return []

    @abstractmethod
    def timestamp(self):
        pass

    @abstractmethod
    def duration(self):
        pass

class TelemetrySignal(Telemetry):
    def __init__(self, message) -> None:
        super().__init__(message)
        self._timestamp = time.perf_counter_ns()
        st = inspect.stack()
        for cf in st:
          f_self = cf.frame.f_locals.get("self", None)
          if f_self is None or not isinstance(f_self, TelemetrySignal):
            break
        self.file = cf.filename
        self.line = cf.lineno
        if 'signals' in globals():
            global signals
            signals.add_signal(self)

    def timestamp(self):
        return self._timestamp

    def duration(self):
        return 0

    def isStartSignal(self):
        return False
    
    def isEndSignal(self):
        return False
    
    @gtView
    def gtViewSignalTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: [self])\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

class TelemetryEvent(Telemetry):
    def __init__(self, message):
        super().__init__(message)
        self._children = []
        self.startSignal = None
        self.endSignal = None

    def children(self):
        return [self.startSignal, *self._children, self.endSignal]

    def addChild(self, child):
        bisect.insort(self._children, child, key=lambda x:x.timestamp())

    def timestamp(self):
        return self.startSignal.timestamp()

    def duration(self):
        return self.endSignal.timestamp() - self.startSignal.timestamp()
    
    def getArgs(self):
        return []
        
    def getKwargs(self):
        return {}
        
    def getResult(self):
        return None
    
    @gtView
    def gtViewEventTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: [ self ])\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

    @gtView
    def gtViewCall(self, aBuilder):
        return aBuilder.columnedList()\
            .title("Call")\
            .priority(5)\
            .items(lambda: [ ("receiver before",self.startSignal.getReceiver()), ("message",self.message), ("arguments",self.startSignal.getArgs()), ("keyword-arguments",self.startSignal.getKwargs()), ("receiver after",self.endSignal.getReceiver()), ("result",self.endSignal.getResult()) ])\
            .column("property", lambda each: each[0])\
            .column("value", lambda each: each[1])

class MethodStartSignal(TelemetrySignal):
    def isStartSignal(self):
        return True
    
    def isEndSignal(self):
        return False

class MethodEndSignal(TelemetrySignal):
    def isStartSignal(self):
        return False
    
    def isEndSignal(self):
        return True

class ArgumentMethodStartSignal(MethodStartSignal):
    def __init__(self, receiver, message, args, kwargs):
        super().__init__(message)
        self.receiver = copy(receiver)
        self.args = args.copy()
        self.kwargs = kwargs.copy()
        
    def getReceiver(self):
        return self.receiver
        
    def getArgs(self):
        return dict(self.args)
        
    def getKwargs(self):
        return self.kwargs
        
    @gtView
    def gtViewCall(self, aBuilder):
        return aBuilder.columnedList()\
            .title("Call")\
            .priority(5)\
            .items(lambda: [ ("receiver",self.receiver), ("message",self.message), ("arguments",self.args), ("keyword-arguments",self.kwargs) ])\
            .column("property", lambda each: each[0])\
            .column("value", lambda each: each[1])
            
class ResultMethodEndSignal(MethodEndSignal):
    def __init__(self, receiver, message, result):
        super().__init__(message)
        self.receiver = copy(receiver)
        self.result = copy(result)
    
    def getReceiver(self):
        return self.receiver
        
    def getResult(self):
        return self.result
    
    @gtView
    def gtViewCall(self, aBuilder):
        return aBuilder.columnedList()\
            .title("Call")\
            .priority(5)\
            .items(lambda: [ ("receiver",self.receiver), ("message",self.message), ("result",self.result) ])\
            .column("property", lambda each: each[0])\
            .column("value", lambda each: each[1])

class TelemetrySignalGroup:
    def __init__(self) -> None:
        self.signals = []

    def get_signals(self):
        return self.signals
    
    def get_event_tree(self):
        b = self.get_signals()
        value = []
        index = 0
        while index < len(b):
            [index, tree] = self.compute_tree(index, b, 0)
            value.append(tree)
        return value
    
    def compute_tree(self, index, list, depth):
        print(f"Depth: {depth}:{index}/{len(list)}")
        if index >= len(list):
            return [index, []]
        if not list[index].isStartSignal():
            return [index+1, list[index]]    # leaf signals
        root = TelemetryEvent(list[index].message)
        root.startSignal = list[index]
        index = index + 1
        while index < len(list) and not list[index].isEndSignal():
            [newindex, kid] = self.compute_tree(index, list, depth+1)
            root.addChild(kid)
            index = newindex
        if index < len(list):
            root.endSignal = list[index]
            index = index + 1
        return [index, root]

    def add_signal(self, signal):
        self.signals.append(signal)

    @gtView
    def gtViewSignals(self, aBuilder):
        return aBuilder.columnedList()\
            .title("Signals")\
            .priority(1)\
            .items(lambda: self.get_signals())\
            .column("Signal Class", lambda each:f"{each.__class__.__name__}")\
            .column("Message", lambda each: each.message)\
            .column("Timestamp", lambda each: each.timestamp())
    
    @gtView
    def gtViewSignalTree(self, aBuilder):
        return aBuilder.columnedTree()\
            .title("Tree")\
            .priority(2)\
            .items(lambda: self.get_event_tree())\
            .children(lambda each: each.children())\
            .column("Message", lambda each: each.message)\
            .column("Duration", lambda each: each.duration())

def start_signals():
    reset_signals()

def reset_signals():
    global signals
    signals = TelemetrySignalGroup()

def get_signals():
    global signals
    return signals



from .PythonBridge.python_bridge import (
	bridge_args, 
	setup_bridge, 
	run_bridge, 
	run_bridge_default, 
	run_bridge_main,
	spawn_bridge,
	stop_spawned_bridge
)

from .gtoolkit.gt import (
	gtView,
	GtViewedObject
)

from .PythonBridge.telemetry import (
	methodevent,
	argmethodevent,
	reset_signals,
	get_signals,
	gtTrace
)

from .__version__ import __version__



import sys
from dataclasses import dataclass
from types import TracebackType


@dataclass
class GtPythonFrame:
    function: str
    filename: str
    lineno: int
    locals: dict


class GtPythonException:
    def __init__(self, exc_type, exc_value, exc_tb: TracebackType):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.exc_tb = exc_tb
        self._frames = self._extract_frames()

    def _safe_locals(self, raw_locals):
        result = {}
        for name, value in raw_locals.items():
            try:
                value_repr = repr(value)
            except Exception:
                value_repr = '<repr failed>'
            result[name] = {
                'type': type(value).__name__,
                'repr': value_repr,
            }
        return result

    def _extract_frames(self):
        frames = []
        tb = self.exc_tb
        while tb:
            frame = tb.tb_frame
            frames.append(
                GtPythonFrame(
                    function=frame.f_code.co_name,
                    filename=frame.f_code.co_filename,
                    lineno=tb.tb_lineno,
                    locals=self._safe_locals(frame.f_locals),
                )
            )
            tb = tb.tb_next
        return frames

    def as_dict(self):
        return {
            'exception_type': self.exc_type.__name__,
            'exception_message': str(self.exc_value),
            'frames': [
                {
                    'function': each.function,
                    'filename': each.filename,
                    'lineno': each.lineno,
                    'locals': each.locals,
                }
                for each in self._frames
            ],
        }
    



from .bridge_globals import bridge_globals

def ensure_global_registry():
    if not hasattr(bridge_globals(), 'obj_registry'):
        bridge_globals()['obj_registry'] = Registry()

def registry():
    return bridge_globals()['obj_registry']

class Registry():

    def __init__(self):
        self.idToRefCount = {}
        self.idToObjMap = {}

    def hasId(self, anId):
        return anId in self.idToObjMap
    
    def register(self, obj):
        if hex(id(obj)) in self.idToObjMap:
            self.idToRefCount[id(obj)] += 1
            return hex(id(obj))
        else:
            return self._register(obj, hex(id(obj)))

    def resolve(self, objId):
        if objId in self.idToObjMap:
            return self.idToObjMap[objId]
        else:
            raise ResolveUnknownObject(objId)

    def _register(self, obj, newObjId):
        self.idToRefCount[id(obj)] = 1
        self.idToObjMap[newObjId] = obj
        return newObjId

    def clean(self, objId):
        obj = self.idToObjMap[objId]
        self.idToRefCount[id(obj)] -= 1
        if self.idToRefCount[id(obj)] == 0:
            del self.idToRefCount[id(obj)]
            del self.idToObjMap[objId]

    def isProxy(self, anObject):
        is_proxy = False

        if isinstance(anObject, dict):
            if len(anObject.keys()) == 2 and ('__pyclass__' in anObject) and ('__pyid__' in anObject):
                is_proxy = True

        return is_proxy

    def proxy(self, obj):
        return {
            '__pyclass__': self.qualifiedNameOf(type(obj)),
            '__pyid__': self.register(obj),
            '__superclasses__': self.superclassesOf(obj)
            }
    
    def superclassesOf(self, obj):
        c = type(obj).__base__
        supers = []
        while c is not None:
            supers.append(self.qualifiedNameOf(c))
            c = c.__base__
        return supers

    def qualifiedNameOf(self, type):
        if type.__module__ is None or type.__module__ == 'builtins':
           return type.__name__
        else:
           return type.__module__ + '.' + type.__name__


class RegistryError(Exception):
    pass

class ResolveUnknownObject(RegistryError):
    def __init__(self, objId):
        RegistryError.__init__(self,"Attempting to resolve unknown object with id {0}.".format(objId))

class RegisterForbiddenObject(RegistryError):
    def __init__(self, obj):
        RegistryError.__init__(self,"Attempting to register forbidden object of type {0}.".format(type(obj).__name__))

ensure_global_registry()



from .view import View
from .data_source import GtPhlowListDataSource

class ListView(View):
	def __init__(self):
		super().__init__()
		self.itemsCallback = lambda : []
		self.itemsFormatCallback = lambda item: str(item)
		self.accessor = None

	def items(self, itemsCallback):
		self.itemsCallback = itemsCallback
		return self
	
	def itemFormat(self, itemFormatCallback):
		self.itemsFormatCallback = itemFormatCallback
		return self

	def dataSource(self):
		return GtPhlowListDataSource(self.itemsCallback, self.itemsFormatCallback, self.accessor)
	
	def asDictionaryForExport(self):
		exportData = super().asDictionaryForExport()
		exportData["viewName"] = "GtPhlowListViewSpecification"
		exportData["dataTransport"] = 2
		return exportData



import argparse
import threading
import time
from gtoolkit_bridge import gtTrace

# Calculate the factorial
@gtTrace
def factorial(n, t):
    time.sleep(n*t/4)
    print("Start: " + str(t) + ": " + str(n))
    if n == 1:
        res = 1
        if t == 1:
            print("Feel free to break here")
    else:
        res = n * factorial(n-1, t)
    return res


# Calculate the factorial and print the result
@gtTrace
def factorial_thread(n, t):
    time.sleep(2)
    result = factorial(n, t)
    print("Thread " + str(t) + " = "+str(result))


def launch_factorials(n):
    threads = []
    print("Calculate: "+str(n))
    breakpoint()
    for i in range(n):
        threads.append(threading.Thread(target=factorial_thread, args=(n+i, i+1)))
        threads[-1].start()
    print("Wait for the results")
    for thread in threads:
        thread.join()
    print("Done")



import msgpack
from .object_registry import registry

primitive_types = [
    type(None), 
    bool, 
    int,
    float,
    bytes,
    str,
    dict,
    list,
    memoryview,
    bytearray]

mapper = {}

def addMapping(key_type, mapping_function):
    mapper[key_type] = mapping_function

class MsgPackSerializer:
    def __init__(self):
        self.primitive_types = primitive_types

    def mapper(self):
        return mapper
    
    def default(self, obj):
        if type(obj) in self.primitive_types:
            return obj
        if type(obj) in self.mapper():
            return self.mapper()[type(obj)](obj)
        return registry().proxy(obj)

    def serialize(self, obj):
        return msgpack.packb(obj, default=self.default, use_bin_type=True)

    def deserialize(self, binary):
        return msgpack.unpackb(binary, raw=False)



from typing import Any

class GtPhlowTextDataSource:
	def __init__(self, func):
		self.func = func

	def getText(self):
		return self.func().asDictionaryForExport()

class GtText:
	def __init__(self, string):
		self.string = string
		self.runGroup = GtPhlowRunGroup()

	def range(self, start, end):
		return GtPhlowTextRun(self, start, end)
	
	def applyAttributes(self, run, attributes):
		self.runGroup.applyAttributes(run, attributes)
	
	def __getattr__(self, name):
		return getattr(self.range(0, len(self.string) - 1), name)

	def asDictionaryForExport(self):
		return { "__typeLabel": "remotePhlowText", "string": self.string, "stylerSpecification": self.runGroup.asDictionaryForExport() }
	
	def gtViewText(self, view):
		text = view.textEditor()
		text.title("Text")
		text.text(lambda: self)
		return text

class GtPhlowRunGroup:
	def __init__(self):
		self.runs = []

	def applyAttributes(self, run, attributes):
		self.runs.append((run, attributes))

	def asDictionaryForExport(self):
		attributedRuns = { "__typeLabel": "phlowRunsGroup", 
			"items": list(map(lambda attributedRun: {
				"__typeLabel": "phlowRun",
				"startIndex": attributedRun[0].start + 1,
				"endIndex": attributedRun[0].end + 1,
				"attributes": list(map(lambda each: each.asDictionaryForExport(), attributedRun[1]))}, self.runs))}
		return { "__typeLabel": "remotePhlowTextAttributeRunsStylerSpecification", "attributeRuns": attributedRuns }

class GtPhlowTextRun:
	def __init__(self, text, start, end):
		self.text = text
		self.start = start
		self.end = end

	def range(self, start, end):
		return self.text.range(start, end)

	def background(self, color):
		self.text.applyAttributes(self, [GtPhlowTextBackgroundAttribute(color)])
		return self

	def black(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("black")])
		return self

	def bold(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("bold")])
		return self
	
	def fontName(self, name):
		self.text.applyAttributes(self, [GtPhlowTextFontNameAttribute(name)])
		return self
	
	def fontSize(self, size):
		self.text.applyAttributes(self, [GtPhlowTextFontSizeAttribute(size)])
		return self

	def foreground(self, color):
		self.text.applyAttributes(self, [GtPhlowTextForegroundAttribute(color)])
		return self
	
	def glamorousCodeFont(self):
		self.fontName("Source Code Pro")
		return self
	
	def glamorousCodeFontAndSize(self):
		self.glamorousCodeFont()
		self.glamorousCodeSize()
		return self
	
	def glamorousCodeFontAndSmallSize(self):
		self.glamorousCodeFont()
		self.glamorousCodeSmallSize()
		return self
	
	def glamorousCodeMiniSize(self):
		self.fontSize(8)
		return self
	
	def glamorousCodeSize(self):
		self.fontSize(14)
		return self
	
	def glamorousCodeSmallSize(self):
		self.fontSize(12)
		return self
	
	def glamorousCodeTinySize(self):
		self.fontSize(10)
		return self
	
	def glamorousFormEditorCodeFontAndSize(self):
		self.glamorousCodeFont()
		self.glamorousCodeTinySize()
		return self
	
	def glamorousMonospace(self):
		return self.glamorousCodeFont()
	
	def glamorousMonospaceBackground(self):
		self.glamorousCodeFontAndSmallSize()
		self.highlight(GtColor.rgb(240, 240, 240))
		return self
	
	def glamorousRegularDefaultFont(self):
		self.fontName("Source Sans Pro")
		return self
	
	def glamorousRegularFont(self):
		self.fontName("Source Sans Pro")
		return self
	
	def glamorousRegularFontAndSize(self):
		self.glamorousRegularFont()
		self.glamorousRegularSize()
		return self
	
	def glamorousRegularSize(self):
		return self.fontSize(14)
	
	def glamorousRegularSmallSize(self):
		return self.fontSize(12)

	def highlight(self, color):
		self.text.applyAttributes(self, [GtPhlowTextHighlightAttribute(color)])
		return self
	
	def italic(self):
		self.text.applyAttributes(self, [GtFontEmphasisAttribute("italic")])
		return self
	
	def light(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("light")])
		return self
	
	def medium(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("medium")])
		return self
	
	def normal(self):
		self.text.applyAttributes(self, [GtFontEmphasisAttribute("normal")])
		return self
	
	def oblique(self):
		self.text.applyAttributes(self, [GtFontEmphasisAttribute("oblique")])
		return self
	
	def regular(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("regular")])
		return self
	
	def thin(self):
		self.text.applyAttributes(self, [GtPhlowTextFontWeightAttribute("thin")])
		return self

class GtPhlowTextAttribute:
	pass

class GtPhlowTextBackgroundAttribute(GtPhlowTextAttribute):
	def __init__(self, color):
		self.color = color

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowTextBackgroundAttribute", "color": self.color.asDictionaryForExport() }

class GtPhlowTextFontWeightAttribute(GtPhlowTextAttribute):
	def __init__(self,weight):
		self.weight = weight

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowFontWeightAttribute", "weight": self.weight }
	
class GtFontEmphasisAttribute(GtPhlowTextAttribute):
	def __init__(self,emphasis):
		self.emphasis = emphasis

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowFontEmphasisAttribute", "emphasis": self.emphasis }

class GtPhlowTextFontNameAttribute(GtPhlowTextAttribute):
	def __init__(self,name):
		self.name = name

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowFontNameAttribute", "name": self.name }

class GtPhlowTextFontSizeAttribute(GtPhlowTextAttribute):
	def __init__(self,size):
		self.size = size

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowFontSizeAttribute", "size": self.size }

class GtPhlowTextForegroundAttribute(GtPhlowTextAttribute):
	def __init__(self, color):
		self.color = color

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowTextForegroundAttribute", "color": self.color.asDictionaryForExport() }

class GtPhlowTextHighlightAttribute(GtPhlowTextAttribute):
	def __init__(self, color):
		self.color = color

	def asDictionaryForExport(self):
		return { "__typeLabel": "phlowTextHighlightAttribute", "color": self.color.asDictionaryForExport() }

class GtPhlowColor:
	pass

class GtPhlowNamedColor(GtPhlowColor):
	def __init__(self, name):
		self.name = name

	def asDictionaryForExport(self):
		return { "name": self.name }

class GtPhlowARGBDColor(GtPhlowColor):
	def __init__(self, r = 0, g = 0, b = 0, a = 255):
		self.r = r
		self.g = g
		self.b = b
		self.a = a

	def asDictionaryForExport(self):
		return { "a": self.a, "r": self.r, "g": self.g, "b": self.b }

class GtColorClass:
	def rgb(self, r, g, b):
		return GtPhlowARGBDColor(r = r, g = g, b = b)
	
	def argb(self, a, r, g, b):
		return GtPhlowARGBDColor(a = a, r = r, g = g, b = b)
	
	def __getattr__(self, name):
		return GtPhlowNamedColor(name)
	
GtColor = GtColorClass()